/*Spiros Ioannou 2/2000 */
/*sivann at softlab.ece.ntua.gr*/

/*
 * changes:
 * 0.6: changed "char ch" to "int ch"
 * 0.7: added -L to link to different file
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <linux/types.h> 
#include <errno.h>

#include <linux/videodev.h>

#define TIMEOUT_LEN 1000
#define MAX_RGBA_IMAGE_SIZE ((704*576*4)+256)


int dev;
int bytes_per_rgb;
struct video_window vid_win;
struct video_capability vid_caps;
struct video_picture vid_pic;
unsigned char pic[MAX_RGBA_IMAGE_SIZE];
int frozen,sequential=0;
unsigned int timeoutid;
time_t thetime;
void save_ppm(char *filename);
void rgb2bgr();
void save_jpeg(char *filename) ;
int jsmooth=0,jopt=0,jquality=60;
int brightness=30,bgr=0,sif=0,vga=0;

void usage() {
  printf("videograb: version 0.5 Spiros Ioannou 2000\n");
  printf("Usage: videograb [OPTION]\n");
  printf("\t-c <num>  number of shots, -1=unlimited (option MANDATORY)\n");
  printf("\t-i <sec>  interval between shots in seconds (default=4)\n");
  printf("\t-d <dev>  device to open (default=/dev/video)\n");
  printf("\t-D <dir>  chdir there\n");
  printf("\t-q        qcif 176x144 output (defaults to cif 352x288)\n");
  printf("\t-F        SIF 320x240 output (for philips pcvc665K and others)\n");
  printf("\t-V        VGA 640x480 output (for philips pcvc665K and others)\n");
  printf("\t-B        RGB2BGR conversion (for philips pcvc665K and others)\n");
  printf("\t-s        use sequential filename numbering (out1,out2...)\n");
  printf("\t-f <str>  filename prefix (used with -s, default=out)\n");
  printf("\t-v        be verbose\n");
  printf("\t-b        run in background as a daemon\n");
  printf("\t-t <int>  set brightness (i.e. 30)     \n");
  printf("\t--------- JPEG OPTIONS -----------------------------------\n");
  printf("\t-j        save output as jpeg (default = ppm)\n");
  printf("\t-O        coding optimization  (default=no)\n");
  printf("\t-Q <num>  quality factor [0-100] (default=60)\n");
  printf("\t-S <num>  smoothing factor [0-100] (default=0)\n");
  printf("\t-l        link last saved file as current0.jpg \n");
  printf("\t-L        link last saved file as current1.jpg \n");
}
int grab ()
{
  static time_t lasttime = 0;
  static int framecount = 0;
  static float rate;
  static double diff = 1.;
  ssize_t len=0,size=0;

  if ( lasttime == 0 ) 
    lasttime = time(NULL);
  
  thetime = time(NULL);
  
  if ((thetime - lasttime) > (int)diff-1) {
    rate = ((float)(framecount)) / diff;
    /*printf("%3.3f fps\n", rate);*/
    framecount = 0;
    lasttime = thetime;
    size = 0;
  };
  
  framecount++;


  errno = 0;
  while( (len = read (dev, pic, MAX_RGBA_IMAGE_SIZE)) <= 0 ) {
    if(errno!=0 && errno!=EINTR) {
      perror("Error read image\n");
      return(0);
    }
  }
  size+=len;
  return (1);
}


void setbrightness(unsigned int value)
{
  vid_pic.brightness = (value)*256;
  if (ioctl (dev, VIDIOCSPICT, &vid_pic) == -1) {
    perror ("ioctl (VIDIOCSPICT)");
    return;
  }
}

void setcolor(unsigned int value)
{
  vid_pic.colour = (value)*256;
  if (ioctl (dev, VIDIOCSPICT, &vid_pic) == -1) {
    perror ("ioctl (VIDIOCSPICT)");
    return;
  }
}

void setcontrast(unsigned int value)
{
  vid_pic.contrast = (value)*256;
  if (ioctl (dev, VIDIOCSPICT, &vid_pic) == -1) {
    perror ("ioctl (VIDIOCSPICT)");
    return;
  }
}

void setsize_vga()
{
  vid_win.x=0;
  vid_win.y=0;
  vid_win.width=640;
  vid_win.height=480;
  if (ioctl (dev, VIDIOCSWIN, &vid_win) == -1) {
    perror ("ioctl (VIDIOCSWIN)");
    return;
  }
  return;
}

void setsize_sif()
{
  vid_win.x=0;
  vid_win.y=0;
  vid_win.width=320;
  vid_win.height=240;
  if (ioctl (dev, VIDIOCSWIN, &vid_win) == -1) {
    perror ("ioctl (VIDIOCSWIN)");
    return;
  }
  return;
}

void setsize_cif()
{
  vid_win.x=0;
  vid_win.y=0;
  vid_win.width=352;
  vid_win.height=288;
  if (ioctl (dev, VIDIOCSWIN, &vid_win) == -1) {
    perror ("ioctl (VIDIOCSWIN)");
    return;
  }
  return;
}


void setsize_qcif()
{
  vid_win.x=88;
  vid_win.y=48;
  vid_win.width=176;
  vid_win.height=144;
  if (ioctl (dev, VIDIOCSWIN, &vid_win) == -1) {
    perror ("ioctl (VIDIOCSWIN)");
    return;
  }
  return; 
}


int main( int   argc, char *argv[] )
{
  int ib=130, ic=30, iw=150;
  unsigned long count=0,j=0;
  int interval=4,qcif=0;
  char filename[128];
  char time_buff[128];
  char hostname[64];
  char device[64];
  char fprefix[64];
  char startdir[128];
  int ch;
  int verbose=0;
  int jpeg=0;
  int bg=0;
  int dolink=0;
  int pid;

  strcpy(device,"/dev/video0");
  strcpy(fprefix,"out");
  startdir[0]=0;

  while ((ch = getopt(argc, argv, "qi:c:d:vsjLOQ:S:f:D:blt:FBV")) != EOF)
     switch ((char)ch) {
     case 'j':
	   jpeg=1; break;
     case 'B':
	   bgr=1; break;
     case 'V':
	   vga=1; break;
     case 'F':
	   sif=1; break;
     case 'S':
	   jsmooth=atoi(optarg); break;
     case 'O':
	   jopt=1; break;
     case 'Q':
	   jquality=atoi(optarg); break;
     case 'b':
	   bg=1; break;
     case 'q':
	   qcif=1; break;
     case 's':
	   sequential=1; break;
     case 't':
	   brightness=atoi(optarg); break;
     case 'i':
	   interval=atoi(optarg); break;
     case 'f':
	   strcpy(fprefix,optarg); break;
     case 'D':
	   strcpy(startdir,optarg); break;
     case 'd':
	   strcpy(device,optarg); break;
     case 'c':
	   count=atol(optarg); break;
     case 'v':
	   verbose=1;break;
     case 'l':
	   dolink=1;break;
     case 'L':
	   dolink=2;break;
     case 'h':
     default:
	usage();
	exit (4);
     }


  if (argc<2 || count==0 ) {
    usage();
    exit (5);
  }

  if (jsmooth<0 || jsmooth>100) jsmooth=0;
  if (jquality<0 || jquality>100) jquality=60;
  if (jopt<0 || jopt>100) jopt=1;

  if (verbose && jpeg)
      printf("smooth:%d, quality:%d, optimize:%d\n",jsmooth,jquality,jopt);

  if (gethostname(hostname,64)) {
    perror("gethostname");
    exit(errno);
  };

  if (startdir[0] && chdir(startdir)) {
    perror(startdir);
    exit (errno);

  }
  dev = open (device, O_RDWR);
  if (dev < 0) {
    perror(device);
    exit(1);
  }

  ioctl (dev, VIDIOCGCAP, &vid_caps);
  ioctl (dev, VIDIOCGWIN, &vid_win);
  if(ioctl (dev, VIDIOCGPICT, &vid_pic)) {
    perror("VIDIOCGPICT");
    exit( 1 );
  } 

  ib = (vid_pic.brightness)/256;
  ic = (vid_pic.contrast)/256;
  iw = (vid_pic.colour)/256;

  vid_pic.palette = VIDEO_PALETTE_RGB24;
  bytes_per_rgb = 3;
  vid_pic.depth = 24;

  /* setbrightness(brightness); */

  if(ioctl (dev, VIDIOCSPICT, &vid_pic)) {
    perror("VIDIOCSPICT");
    exit( 1 );
  } 

  vid_win.x=0;
  vid_win.y=0;
  vid_win.width=352;
  vid_win.height=288;

  if (ioctl (dev, VIDIOCSWIN, &vid_win) == -1) {
    perror ("ioctl (VIDIOCSWIN)");
    exit( 1 );
  }
    
  if (qcif)
    setsize_qcif();
  else if (sif)
    setsize_sif();
  else if (vga)
    setsize_vga();
  else
    setsize_cif();

  if ( bg && (pid=fork())) {
    if (verbose) printf("forking\n");
    exit (0);
  }

  for (j=0;j<count||count<0;j++) {
    if (count<0) j=0;
    if (verbose) {
      printf("iteration:%ld\n",j);
      printf("count:%ld\n",count);
      printf("sequential:%d\n",sequential);
      printf("sleeping for: %d seconds\n",interval);
    }
    sleep(interval);
    if (verbose) 
      printf("grabbing...");fflush(stdout);
    grab();
    if (verbose) 
      printf("ok\n");


    strftime(time_buff,128,"%Y%j-%T",localtime(&thetime));
    sprintf(filename,"%s-spy-%s",hostname,time_buff);
    if (sequential) sprintf(filename,"%s%02ld",fprefix,j);

    if (bgr)
      rgb2bgr();

    if (jpeg) {
      strcat(filename,".jpg");
      if (verbose) printf("Writting %s\n",filename);
      save_jpeg(filename);
      if (dolink==1) {
	unlink("current0.jpg");
	if (link(filename,"current0.jpg")) {
	  perror("link:current0.jpg");
	}
      }
      else if (dolink==2) {
	unlink("current1.jpg");
	if (link(filename,"current1.jpg")) {
	  perror("link:current1.jpg");
	}
      }
    }
    else {
      strcat(filename,".ppm");
      if (verbose) printf("Writting %s\n",filename);
      save_ppm(filename);
    }
  }
  return 0;
}

//convert pic[] to bgr
void rgb2bgr()
{
  int i;
  unsigned char b1;

  for (i = 0; i < vid_win.width*vid_win.height*3; i+=3) {
    b1 = pic[i];   //R
    pic[i]=pic[i+2];
    pic[i+2]=b1;
  }
 
}

void save_ppm(char *filename)
{
  FILE * outfile;
  unsigned char buff[3];
  int i;

    if (!(outfile=fopen(filename,"w"))) {
      perror(filename);
      exit(errno);
    };
    fprintf (outfile, "P6\n%d %d\n%d\n", vid_win.width, vid_win.height, 255);
    for (i = 0; i < vid_win.width*vid_win.height*3; i+=3) {
      buff[0] = pic[i];   //R
      buff[1] = pic[i+1]; //G
      buff[2] = pic[i+2]; //B
      fwrite (buff, 1, 3, outfile);
    }
    fclose(outfile);
}


#ifdef JPEG
#include <jpeglib.h>
void save_jpeg(char *filename)
{
FILE *outfile;
struct jpeg_compress_struct cinfo;
struct jpeg_error_mgr jerr;
int row_stride;              // physical row width in image buffer
JSAMPROW row_pointer[1];     // pointer to JSAMPLE row(s)

  if (!(outfile=fopen(filename,"w"))) {
    perror(filename);
    exit(errno);
  };

  cinfo.err = jpeg_std_error(&jerr);
  jpeg_create_compress(&cinfo);
  
  jpeg_stdio_dest(&cinfo, outfile);
  cinfo.image_width = vid_win.width;  // image width and height, in pixels
  cinfo.image_height = vid_win.height;
  cinfo.input_components = 3;         // # of color components per pixel
  cinfo.in_color_space = JCS_RGB;     // colorspace of input image

  cinfo.smoothing_factor = jsmooth;
  cinfo.optimize_coding = jopt;
  
  jpeg_set_defaults(&cinfo);
  jpeg_set_quality(&cinfo,jquality,TRUE);
  jpeg_start_compress(&cinfo, TRUE);
  row_stride=vid_win.width*3;
  while (cinfo.next_scanline < cinfo.image_height) {
    /* jpeg_write_scanlines expects an array of pointers to scanlines.
     * Here the array is only one element long, but you could pass
     * more than one scanline at a time if that's more convenient.
     */
    row_pointer[0] = &pic[cinfo.next_scanline * row_stride];
    jpeg_write_scanlines(&cinfo, row_pointer, 1);
  }
  jpeg_finish_compress(&cinfo);
  jpeg_destroy_compress(&cinfo);
  fclose(outfile);
}

#else
void save_jpeg(char *filename) {
  fprintf(stderr,"JPEG support was not compiled in.\n");
  fprintf(stderr,"Please edit the Makefile and recompile.\n");
  exit(0);
}
#endif
